/*----------------------------------------------------------------------------
    TAPESTREA: Techniques And Paradigms for Expressive Synthesis, 
               Transformation, and Rendering of Environmental Audio
      Engine and User Interface

    Copyright (c) 2006 Ananya Misra, Perry R. Cook, and Ge Wang.
      http://taps.cs.princeton.edu/
      http://soundlab.cs.princeton.edu/

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
    U.S.A.
-----------------------------------------------------------------------------*/

//-----------------------------------------------------------------------------
// name: ui_main.cpp
// desc: entry point for taps
//
// authors: Ananya Misra (amisra@cs.princeton.edu)
//          Philip Davidson (philipd@cs.princeton.edu)
//          Ge Wang (gewang@cs.princeton.edu)
//          Perry R. Cook (prc@cs.princeton.edu)
// date: Spring 2006
//-----------------------------------------------------------------------------
#include "audicle_def.h"
#include "audicle_gfx.h"
#include "audicle.h"

#include "ui_audio.h"
#include "ui_analysis.h"
#include "ui_treesynth.h"
#include "ui_synthesis.h"
#include "ui_control.h"
#include "ui_library.h"
#include "ui_search.h"
#include "audicle_elcidua.h"

#include <stdlib.h>
#include <time.h>


// version
#define TAPS_VERSION "0.1.0.0"

// function prototype
void print_usage();

#ifdef __MACOSX_CORE__
#include <sys/types.h>
#include <sys/time.h>
#include <sys/resource.h>

// free
void osx_increase_file_handles()
{
    struct rlimit limit;
    int err;

    err = getrlimit(RLIMIT_NOFILE, &limit);
    if (err == 0) {
        limit.rlim_cur = RLIM_INFINITY;
        (void) setrlimit(RLIMIT_NOFILE, &limit);
    }
}
#endif

#ifdef __PLATFORM_LINUX__
#include <gtk/gtk.h>

int g_argc;
char ** g_argv;

// gtk thread
void * cb_gtk( void * )
{
    // init
    gtk_init( &g_argc, &g_argv );
    // just run it
    gtk_main();
    return NULL;
}

CHUCK_THREAD g_gtk_id;

#endif


//-----------------------------------------------------------------------------
// name: main()
// desc: entry point
//-----------------------------------------------------------------------------
int main( int argc, char ** argv )
{
    // you know
    srand( time( NULL ) );
    // info
    BB_log( BB_LOG_SYSTEM, "Tapestrea version: %s", TAPS_VERSION );
    BB_setlog( BB_LOG_INFO );
    
    // fullscreen
    t_CKBOOL fullscreen = TRUE;

#ifdef __MACOSX_CORE__
    // limits
    osx_increase_file_handles();
    // sample rate
    BirdBrain::our_srate = 44100;
    // no
    fullscreen = TRUE;
#endif

#ifdef __PLATFORM_WIN32__
    // sample rate
    BirdBrain::our_srate = 44100;
#endif

#ifdef __PLATFORM_LINUX__
    fullscreen = FALSE;
    // sample rate
    BirdBrain::our_srate = 48000;
    // copy
    g_argc = argc;
    g_argv = argv;
    // init
    gtk_init( &g_argc, &g_argv );
    // go
    // pthread_create( &g_gtk_id, NULL, cb_gtk, NULL );
#endif

    // parse arguments
    int index = 1;
    while( index < argc )
    {               
        // sample rate
        if( !strcmp( argv[index], "--srate" ) )
        {
            if( ++index < argc )
            {
                BirdBrain::our_srate = atoi( argv[index] );
                BB_log( BB_LOG_INFO, "(main) Setting sample rate to %i Hz", BirdBrain::srate() );
            }
        }
        // preprocess mode (always the last thing)
        else if( !strcmp( argv[index], "--preprocess" ) )
        {
            BB_log( BB_LOG_SYSTEM, "Preprocess mode" ); 
            Driver jill;
            jill.set( 512, 4096, 1 );
            BirdBrain::our_hop_size = 512 / 4;  
            while( ++index < argc )
            {
                if( !jill.open( argv[index] ) )
                {
                    BB_log( BB_LOG_SYSTEM, "(main) Driver could not open file %s", argv[index] );
                    continue;
                }
                
                jill.preprocess(); 
                jill.brake(); 
            }

            return 0;
        }
        // log level
        else 
            BB_setlog( atoi( argv[index] ) ); 

        index++; 
    }

    // normal mode
    // start audio
    AudioCentral::instance()->init( BirdBrain::srate(), 1024, 128 );
    // initialize graphics
    AudicleGfx::init();
    // initialize main window
    AudicleWindow::main()->init( 1024, 768, 0, 0, "tapestrea", fullscreen );

    // launch audicle
    Audicle::instance()->init();
    // add faces to the audicle
    Audicle::instance()->add( new UIAnalysis );
    Audicle::instance()->add( new UISynthesis );
    Audicle::instance()->add( new UIControl );
    Audicle::instance()->add( new UISearch );
    Audicle::instance()->add( new ElciduaFace );
    Audicle::instance()->add( new ElciduaFace );
    //Audicle::instance()->add( new ElciduaFace );
    // Audicle::instance()->add( new UITreesynth );
    // move to the initial face
    Audicle::instance()->move_to( (t_CKUINT)0 );

    BB_log( BB_LOG_SYSTEM, "TAPESTREA system ready..." );
    
    // print usage
    print_usage();

    // main loop
    AudicleGfx::loop();

    return 0;
}




//-----------------------------------------------------------------------------
// name: print_usage()
// desc: print usage
//-----------------------------------------------------------------------------
void print_usage()
{
    fprintf( stderr, "----------------------------------------------------\n" );
    fprintf( stderr, "Tapestrea (1.0.0.0)\n" );
    fprintf( stderr, "----------------------------------------------------\n" );
    fprintf( stderr, "'Ctrl + arrow' - switch faces\n" );
    fprintf( stderr, "'Ctrl + g' - toggle full screen\n" );
    fprintf( stderr, "'Ctrl + q' or 'Apple + q' - quit\n" );
    fprintf( stderr, "'up or down arrow' - change volume\n" );
    fprintf( stderr, "'h' - toggle highlighting\n" );
    fprintf( stderr, "'1', '2', '3', ... - toggle zoom on different regions\n" );
    fprintf( stderr, "'0', 'esc' - zoom out\n" );
    fprintf( stderr, "'r' - toggle display of slider ranges\n" );
    fprintf( stderr, "----------------Analysis---------------------------\n" );
    fprintf( stderr, "'L' - toggle spectrogram selection rectangle line width\n" );
    fprintf( stderr, "----------------Synthesis--------------------------\n" );
    fprintf( stderr, "'Backspace' - delete selected template from library\n" );
    fprintf( stderr, "'N' - create new mixed bag\n" );
    fprintf( stderr, "'q' or 'z' - toggle display of quantization markers\n" );
    fprintf( stderr, "Right clicking on a template plays/stops it\n" );
    fprintf( stderr, "Drag to move templates onto/off/within a timeline\n" );
    fprintf( stderr, "-----------------Mixer Face-------------------------\n" );
    fprintf( stderr, "'Ctrl + down arrow' - to access\n" );
    fprintf( stderr, "Bus 0 - analysis\n" );
    fprintf( stderr, "Bus 2 - synthesis\n" );
    fprintf( stderr, "Bus 7 - chuck scripts (default)\n" );
    fprintf( stderr, "----------------Treesynth Face---------------------\n" );
    fprintf( stderr, "'Ctrl + any arrow' - leave at once\n" );
    fprintf( stderr, "----------------No face (preprocess)---------------\n" );
    fprintf( stderr, "'taps --preprocess file1.wav file2.wav ... fileN.wav'\n" );
    fprintf( stderr, "    preprocess each wav file\n" );
    fprintf( stderr, "    outputs: file1.pp, file1.fft, file2.pp, file2.fft, ...\n" );
    fprintf( stderr, "    later in analysis face, load the .pp file\n" );
    fprintf( stderr, "----------------No face (sample rate)---------------\n" );
    fprintf( stderr, "'taps --srate s'\n" );
    fprintf( stderr, "    set working sample rate to s Hz\n" );
    fprintf( stderr, "----------------------------------------------------\n" );
}
